#!/bin/env python

import cython
from cython.cimports.tadashi import isl
from cython.cimports.tadashi.scop import Scop
from cython.cimports.tadashi.transformations import *

@cython.ccall
def tile_1d(scop: Scop, tile_size: int):
    scop.ptr_ccscop.current_node = tadashi_tile_1d(scop.ptr_ccscop.current_node, tile_size)

@cython.ccall
def tile_2d(scop: Scop, size1: int, size2: int):
    scop.ptr_ccscop.current_node = tadashi_tile_2d(scop.ptr_ccscop.current_node, size1, size2)

@cython.ccall
def tile_3d(scop: Scop, size1: int, size2: int, size3: int):
    scop.ptr_ccscop.current_node = tadashi_tile_3d(scop.ptr_ccscop.current_node, size1, size2, size3)

@cython.ccall
def interchange(scop: Scop):
    scop.ptr_ccscop.current_node = tadashi_interchange(scop.ptr_ccscop.current_node)

@cython.ccall
def full_fuse(scop: Scop):
    scop.ptr_ccscop.current_node = tadashi_full_fuse(scop.ptr_ccscop.current_node)

@cython.ccall
def fuse(scop: Scop, idx1: int, idx2: int):
    scop.ptr_ccscop.current_node = tadashi_fuse(scop.ptr_ccscop.current_node, idx1, idx2)

@cython.ccall
def full_split(scop: Scop):
    scop.ptr_ccscop.current_node = tadashi_full_split(scop.ptr_ccscop.current_node)

@cython.ccall
def split(scop: Scop, split_idx: int):
    scop.ptr_ccscop.current_node = tadashi_split(scop.ptr_ccscop.current_node, split_idx)

@cython.ccall
def scale(scop: Scop, scale: int):
    scop.ptr_ccscop.current_node = tadashi_scale(scop.ptr_ccscop.current_node, scale)

@cython.ccall
def full_shift_val(scop: Scop, val: int):
    scop.ptr_ccscop.current_node = tadashi_full_shift_val(scop.ptr_ccscop.current_node, val)

@cython.ccall
def partial_shift_val(scop: Scop, pa_idx: int, val: int):
    scop.ptr_ccscop.current_node = tadashi_partial_shift_val(scop.ptr_ccscop.current_node, pa_idx, val)

@cython.ccall
def full_shift_var(scop: Scop, var_idx: int, coeff: int):
    scop.ptr_ccscop.current_node = tadashi_full_shift_var(scop.ptr_ccscop.current_node, var_idx, coeff)

@cython.ccall
def partial_shift_var(scop: Scop, pa_idx: int, id_idx: int, coeff: int):
    scop.ptr_ccscop.current_node = tadashi_partial_shift_var(scop.ptr_ccscop.current_node, pa_idx, id_idx, coeff)

@cython.ccall
def full_shift_param(scop: Scop, param_idx: int, coeff: int):
    scop.ptr_ccscop.current_node = tadashi_full_shift_param(scop.ptr_ccscop.current_node, param_idx, coeff)

@cython.ccall
def partial_shift_param(scop: Scop, pa_idx: int, param_idx: int, coeff: int):
    scop.ptr_ccscop.current_node = tadashi_partial_shift_param(scop.ptr_ccscop.current_node, pa_idx, param_idx, coeff)

@cython.ccall
def set_parallel(scop: Scop, num_threads: int):
    scop.ptr_ccscop.current_node = tadashi_set_parallel(scop.ptr_ccscop.current_node, num_threads)

@cython.ccall
def set_loop_opt(scop: Scop, pos: int, opt: int):
    scop.ptr_ccscop.current_node = tadashi_set_loop_opt(scop.ptr_ccscop.current_node, pos, opt)




#!/usr/bin/env python
import argparse
import difflib
import json
import random
import subprocess
import time
from pathlib import Path
from subprocess import TimeoutExpired

from tadashi import Scop
from tadashi.apps import Polybench, Simple


def get_args():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--verify",
        default=False,
        help="Verify instead of measure",
        action="store_true",
    )
    parser.add_argument(
        "--seed",
        type=int,
        default=2,
        help="Random seed",
    )
    parser.add_argument(
        "--num-steps",
        type=int,
        default=1,
        help="Number of transformations",
    )
    return parser.parse_args()


class Model:
    def __init__(self):
        self.node_idx = 0

    def random_transform(self, scop: Scop):
        node = scop.schedule_tree[self.node_idx]
        indices = [node.parent_idx] if self.node_idx > 0 else []
        self.node_idx = random.choice(indices + node.children_idx)
        node = scop.schedule_tree[self.node_idx]
        while not node.available_transformations:
            indices = [node.parent_idx] if self.node_idx > 0 else []
            self.node_idx = random.choice(indices + node.children_idx)
            node = scop.schedule_tree[self.node_idx]
        tr = random.choice(node.available_transformations)
        args = node.get_args(tr, start=-64, end=64)
        return [node.index, tr, *random.choice(args)]


class Timer:
    def __init__(self):
        self.times = [{}]
        self.tick = time.monotonic()

    def time(self, key):
        t = time.monotonic()
        self.times[-1][key] = t - self.tick
        self.tick = t
        return t

    def reset(self):
        self.tick = time.monotonic()

    def custom(self, key, value):
        self.times[-1][key] = value


def get_array(app: Polybench):
    result = subprocess.run(
        app.run_cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        timeout=40,
    )
    output = result.stderr.decode()
    return output.split("\n")


def run_model(app, num_steps, name=""):
    model = Model()
    timer = Timer()
    app.compile()
    timer.time("Compilation")
    app.measure()
    t = timer.time("Total walltime")
    scop = app.scops[0]
    timer.time("Extraction")
    timer.custom("Kernel walltime", t)
    trs = []
    for i in range(num_steps):
        timer.times.append({})
        timer.reset()
        timer.time("Random transformation")
        scop.reset()
        scop.transform_list(trs)
        tr = model.random_transform(scop)
        legal = scop.transform_list([tr])[-1]
        if legal:
            trs.append(tr)
        timer.time("Transformation + legality")
    scop.reset()
    scop.transform_list(trs)
    timer.reset()
    print(f"{trs=}")
    print("transformed")
    tapp = app.generate_code()
    print("code generated")
    timer.time("Code generation")
    tapp.compile()
    print("compiled")
    timer.time("Compilation")
    try:
        t = tapp.measure(timeout=10)
        timer.time("Total walltime")
        timer.custom("Kernel walltime", t)
    except TimeoutExpired as e:
        print(f"Timeout expired: {e=}")
    times_dir = Path("./times")
    times_dir.mkdir(exist_ok=True)
    filename = times_dir / f"{name}-{num_steps}.json"
    json.dump(timer.times, open(filename, "w"))
    print(f"Written: {filename}")


def measure_polybench(num_steps):
    for i, p in enumerate(Polybench.get_benchmarks()):
        print(f"Start {i+1}. {p.name}")
        app = Polybench(p, compiler_options=["-DEXTRALARGE_DATASET", "-O3"])
        print("app done")
        run_model(app, num_steps=num_steps, name=p.name)


def verify_polybench():
    compiler_options = ["-DSMALL_DATASET", "-DPOLYBENCH_DUMP_ARRAYS"]
    for p in Polybench.get_benchmarks():
        app = Polybench(p, compiler_options=compiler_options)
        app.compile()
        gold = get_array(app)
        # print(f"{gold[:3]=}")
        run_model(app, num_steps=3, name=p.name)
        # app.compile()
        mod = get_array(app)
        # print(f"{mod [:3]=}")
        diff = "\n".join(difflib.unified_diff(gold, mod))
        if diff:
            print("<<<<<<<<<<<<< ERROR")
            print(diff)
        else:
            print(f">>>>>>>>>>>>> OK {p}")


if __name__ == "__main__":
    args = get_args()
    random.seed(args.seed)
    if args.verify:
        verify_polybench()
    else:
        measure_polybench(num_steps=args.num_steps)



from tadashi import TrEnum

#from .optimize import opimize_app

# TODO: make this proper config
allowed_transformations = {
    TrEnum.TILE1D,
    TrEnum.TILE2D,
    TrEnum.TILE3D,
    TrEnum.INTERCHANGE,
    TrEnum.FULL_FUSE,
    TrEnum.FULL_SPLIT,
}


config = {"cnt_rollouts": 0,
          "repeats": 1,
          "rollouts": 1,
          "cnt_evals": 0,
          "scop_idx": 0,
          "max_depth": 2,
          #"whitelist_transformations": allowed_transformations
          }




#!/bin/env python

import tempfile
import unittest
from pathlib import Path

from tadashi.scop import Scop
from tadashi.translators import *


class TestTranslator(unittest.TestCase):
    """!!! Consider other functionality. Also consider testing ccscop
    functionality.

    """

    examples: Path = Path(__file__).parent.parent / "examples"
    tests: Path = Path(__file__).parent

    def test_pet_autodetect(self):
        """Test the `translators.Pet`'s `autodetect` parameter."""
        file = self.examples / "inputs/depnodep.c"
        cases = [
            (False, 1),
            (True, 3),
        ]
        for autodetect, num_scops in cases:
            with self.subTest(autodetect=autodetect):
                translator = Pet(autodetect=autodetect)
                translator.set_source(file, [])
                self.assertEqual(len(translator.scops), num_scops)

    def test_double_set(self):
        """Calling set_source() 2x on a translator should raise an error!"""
        translator = Pet()
        file = self.examples / "inputs/depnodep.c"
        translator.set_source(file, [])
        with self.assertRaises(RuntimeError):
            translator.set_source(file, [])

    def test_deleted_translator(self):
        """This test should simply not crash."""

        def _get_schedule_tree():
            file = self.examples / "inputs/depnodep.c"
            translator = Pet()
            translator.set_source(file, [])
            scop = translator.scops[0]
            # `translator` is freed after the next line.
            return scop.schedule_tree

        with self.assertRaises(RuntimeError):
            node = _get_schedule_tree()[1]
            # The translator is qeried when in the next line!
            print(node.yaml_str)

    @unittest.skip("todo")
    def test_pet_scop_extraction(self):
        """Test the things done in the constructor: does it populate
        ccscops and scops correctly."""

    @unittest.skip("todo")
    def test_pet_codegen(self):
        """Test codegen"""

    @unittest.skip("todo")
    def test_polly_codegen(self):
        """See same test for pet."""

    @unittest.skip("todo")
    def test_polly_scop_extraction(self):
        """See same test for pet."""


class TestPet(TestTranslator):
    def test_non_existing_file(self):
        """Test if Pet errors out when we try to open an non-existing file."""
        with tempfile.TemporaryDirectory() as tmp:
            print(f"{tmp=} {type(tmp)=}")
            path = Path(tmp) / "does-not-exists.c"
            self.assertFalse(path.exists())
            with self.assertRaises(ValueError):
                translator = Pet()
                translator.set_source(path, [])

    def test_compilation_error(self):
        with self.assertRaises(ValueError):
            translator = Pet(autodetect=True)
            path = self.tests / "syntax_error.c"
            self.assertTrue(path.exists())
            translator.set_source(path, [])



#!/bin/env python
import multiprocessing
import re
from ast import literal_eval
from collections import namedtuple
from dataclasses import dataclass
from enum import Enum, StrEnum, auto
from typing import Tuple

import cython
from cython.cimports.tadashi import isl, pet, transformations
from cython.cimports.tadashi.ccscop import ccScop

from . import _tr_wrappers as w
from .node_type import NodeType

SHOULD_NOT_HAPPEN = "This should not have happened. Please report an issue at https://github.com/vatai/tadashi/issues/"
DELETED_TRANSLATOR_EXCMSG = (
    "Something is wrong! Was the translator deleted before the scop?"
)


class AstLoopType(Enum):
    """Possible values for `SET_LOOP_OPT`.

    `UNROLL` should be avoided unless the requirements in the
    `ISL Docs <https://libisl.sourceforge.io/user.html#AST-Generation-Options-Schedule-Tree>`_.
    are satisfied.

    """

    DEFAULT = 0
    ATOMIC = auto()
    UNROLL = auto()
    SEPARATE = auto()

    def __repr__(self):
        return f"TrEnum.{self.value.upper()}"


LowerUpperBound = namedtuple(
    "LowerUpperBound",
    ["lower", "upper"],
    defaults=[None, None],
)
"""Integer interval description.

Lower and upper bounds for describing (integer) intervals of valid
arguments for transformations. `None` indicates no upper/lower bound.

"""


CAMEL_TO_SNAKE = re.compile(r"(?<=[a-z])(?=[A-Z0-9])")


class TrEnum(StrEnum):
    """Enums of implemented transformations.

    One of these enums needs to be passed to `Node.transform()` (with
    args) to perform the transformation.
    """

    TILE_1D = auto()
    TILE_2D = auto()
    TILE_3D = auto()
    INTERCHANGE = auto()
    FULL_FUSE = auto()
    FUSE = auto()
    FULL_SPLIT = auto()
    SPLIT = auto()
    SCALE = auto()
    FULL_SHIFT_VAL = auto()
    PARTIAL_SHIFT_VAL = auto()
    FULL_SHIFT_VAR = auto()
    PARTIAL_SHIFT_VAR = auto()
    FULL_SHIFT_PARAM = auto()
    PARTIAL_SHIFT_PARAM = auto()
    SET_PARALLEL = auto()
    SET_LOOP_OPT = auto()


@cython.cclass
class Node:
    """Schedule node (Python representation)."""

    scop: Scop
    _type = cython.declare(isl.isl_schedule_node_type, visibility="public")
    num_children = cython.declare(int, visibility="public")
    parent_idx = cython.declare(int, visibility="public")
    index = cython.declare(int, visibility="public")
    label = cython.declare(str, visibility="public")
    location = cython.declare(list[int], visibility="public")
    loop_signature = cython.declare(list, visibility="public")
    expr = cython.declare(str, visibility="public")
    children_idx = cython.declare(list[str], visibility="public")

    def transform(self, tr: TrEnum, *args) -> bool:
        """Execute the selected transformation.

        Args:
          TrEnum tr: Transformation `Enum`.
          args: Arguments passed to the transformation corresponding toe `trkey`.
        """

        nargs = len(args)
        target = TRANSFORMATIONS[tr].num_args()
        if nargs != target:
            raise ValueError(f"Number of args ({nargs}) for {tr} should be {target}")
        if not TRANSFORMATIONS[tr].valid(self):
            raise ValueError(f"Transformation {tr} not valied here:\n{self.yaml_str}")
        if not TRANSFORMATIONS[tr].valid_args(self, *args):
            tr_name = tr.__class__.__name__
            msg = f"Not valid {args=}, for {tr_name=}"
            raise ValueError(msg)
        # TODO proc_args (from olden times)
        self.scop._locate(self.location)
        ps = self.scop.ptr_ccscop
        if ps.tmp_node != cython.NULL:
            isl.isl_schedule_node_free(ps.tmp_node)
        ps.tmp_node = isl.isl_schedule_node_copy(ps.current_node)
        ps.tmp_legal = ps.current_legal
        TRANSFORMATIONS[tr].transform(self.scop, *args)
        ps.current_legal = ps.check_legality()
        ps.modified = True
        return bool(self.scop.ptr_ccscop.current_legal)

    @property
    def node_type(self) -> NodeType:
        return NodeType(self._type)

    @staticmethod
    @cython.cfunc
    def create(
        ptr: cython.pointer(ccScop),
        parent: int,
        current_idx: int,
        location: list[int],
    ) -> Node:
        num_children = isl.isl_schedule_node_n_children(ptr.current_node)
        node_type = isl.isl_schedule_node_get_type(ptr.current_node)
        node = Node()
        scop = Scop.create(ptr)
        node.scop = scop
        node._type = node_type
        node.num_children = num_children
        node.parent_idx = parent
        node.index = current_idx
        node.label = scop._get_label()
        node.location = location
        node.loop_signature = literal_eval(scop._get_loop_signature())
        node.expr = scop._get_expr()
        node.children_idx = [-1] * num_children
        return node

    @property
    def parent(self):
        """The node which is the parent of the current node."""
        return self.scop.schedule_tree[self.parent_idx]

    @property
    def children(self):
        """List of nodes which are the children of the current node."""
        return [self.scop.schedule_tree[i] for i in self.children_idx]

    def available_args(self, tr: TrEnum) -> list:
        """Describe available args."""
        return TRANSFORMATIONS[tr].available_args(self)

    def get_args(self, tr: TrEnum, start: int, end: int) -> list:
        expanded: list[list] = [[]]
        params = self.available_args(tr)
        if not params:
            return expanded
        if all([isinstance(p, list) for p in params]):
            return params
        for param in params:
            args = param
            if isinstance(param, LowerUpperBound):
                if param.lower is not None and param.lower > start:
                    start = param.lower
                if param.upper is not None and param.upper < end:
                    end = param.upper
                args = list(range(start, end))
                expanded = [[*e, a] for e in expanded for a in args]
            elif isinstance(param, list) and len(param) and isinstance(param[0], list):
                expanded = [[*e, *a] for e in expanded for a in args]
            else:
                expanded = [[*e, a] for e in expanded for a in args]
        return expanded

    def valid_transformation(self, tr: TrEnum) -> bool:  # todo add test
        """Check the validity of the transformation."""
        return TRANSFORMATIONS[tr].valid(self)

    @property  # todo add test
    def available_transformations(self) -> list[TrEnum]:
        """List transformations available at the `Node`."""
        result = []
        for k, tr in TRANSFORMATIONS.items():
            if tr.valid(self):
                result.append(k)
        return result

    def valid_args(self, tr: TrEnum, *args) -> bool:  # todo add test
        """Check the validity of args."""
        return TRANSFORMATIONS[tr].valid_args(self, *args)

    @property
    def yaml_str(self):
        self.scop._locate(self.location)
        return self.scop._yaml_str()

    def __repr__(self):
        """Textual representation of a node."""
        words = [
            "Node type:",
            f"{self.node_type},",
            f"{self.loop_signature},",
            f"{self.expr},",
            f"{self.location}",
        ]
        return " ".join(words)


TRANSFORMATIONS = {}


def register(cls):
    """Register classes to TRANSFORMATIONS dict."""

    funcname = cls.__name__
    funcname = funcname.replace("Info", "")
    funcname = CAMEL_TO_SNAKE.sub("_", funcname).lower()
    cls.transform = getattr(w, funcname)
    TRANSFORMATIONS[TrEnum(funcname)] = cls
    return cls


class TransformInfo:
    """Abstract base class used to describe transformations."""

    arg_help: list[str] = []
    """Help string describing the arg."""

    @staticmethod
    def valid(node: Node) -> bool:
        """Check that the transformation is valid on the node."""
        return node.node_type == NodeType.BAND

    @staticmethod
    def valid_args(node: Node, *arg, **kwargs) -> bool:
        """Check that args of the transformation is valid on node."""
        return True

    @staticmethod
    def _is_valid_stmt_idx(node: Node, idx: int) -> bool:
        return 0 <= idx < len(node.loop_signature)

    @staticmethod
    def _valid_idx_all_stmt(node: Node, idx: int, stmt_key: str) -> bool:
        return all(0 <= idx < len(s[stmt_key]) for s in node.loop_signature)

    @staticmethod
    def _is_valid_child_idx(node: Node, idx):
        return 0 <= idx and idx < len(node.children)

    @staticmethod
    def available_args(node: Node) -> list:
        """Return a list describing each of the args."""
        return []

    @classmethod
    def num_args(cls):
        return len(cls.arg_help)


def _tilable(node: Node, dim: int) -> bool:  # todo: try to move to Node
    for _ in range(dim):
        if node.node_type != NodeType.BAND:
            return False
        # if "tile" in node.label and "outer" in node.label:
        #     return False
        node = node.children[0]
    return True


@register
class Tile1DInfo(TransformInfo):
    arg_help = ["Tile size"]

    @staticmethod
    def valid(node: Node):
        return _tilable(node, 1)

    @staticmethod
    def valid_args(node, size1):
        return size1 > 1

    @staticmethod
    def available_args(node: Node):
        return [
            LowerUpperBound(lower=2, upper=None),
        ]


@register
class Tile2DInfo(TransformInfo):
    arg_help = ["Size1", "Size2"]

    @staticmethod
    def valid(node: Node):
        return _tilable(node, 2)

    @staticmethod
    def valid_args(node, size1, size2):
        return size1 > 1 and size2 > 1

    @staticmethod
    def available_args(node: Node):
        return [
            LowerUpperBound(lower=2, upper=None),
            LowerUpperBound(lower=2, upper=None),
        ]


@register
class Tile3DInfo(TransformInfo):
    arg_help = ["Size1", "Size2", "Size3"]

    @staticmethod
    def valid(node: Node):
        return _tilable(node, 3)

    @staticmethod
    def valid_args(node, size1, size2, size3):
        return size1 > 0 and size2 > 0 and size3 > 0

    @staticmethod
    def available_args(node: Node):
        return [
            LowerUpperBound(lower=2, upper=None),
            LowerUpperBound(lower=2, upper=None),
            LowerUpperBound(lower=2, upper=None),
        ]


@register
class InterchangeInfo(TransformInfo):
    @staticmethod
    def valid(node: Node):
        return (
            node.node_type == NodeType.BAND
            and len(node.children) == 1
            and node.children[0].node_type == NodeType.BAND
        )


@register
class FullFuseInfo(TransformInfo):
    @staticmethod
    def valid(node: Node):
        return (
            node.node_type == NodeType.SEQUENCE or node.node_type == NodeType.SET
        ) and all(ch.children[0].node_type == NodeType.BAND for ch in node.children)


@register
class FuseInfo(TransformInfo):
    arg_help = ["Index of first loop to fuse", "Index of second loop to fuse"]

    @staticmethod
    def valid(node: Node):
        return (
            (node.node_type == NodeType.SEQUENCE or node.node_type == NodeType.SET)
            and (len(node.children) > 1)
            and (all(ch.children[0].node_type == NodeType.BAND for ch in node.children))
        )

    @staticmethod
    def valid_args(node: Node, loop_idx1: int, loop_idx2: int):
        return (
            TransformInfo._is_valid_child_idx(node, loop_idx1)
            and TransformInfo._is_valid_child_idx(node, loop_idx2),
        )

    @staticmethod
    def available_args(node: Node):
        nc = len(node.children)
        args = []
        for arg1 in range(nc):
            for arg2 in range(arg1 + 1, nc):
                args.append([arg1, arg2])
        return args


@register
class FullSplitInfo(TransformInfo):
    # TODO -> split sequence!
    @staticmethod
    def valid(node: Node):
        if node.node_type not in [NodeType.SEQUENCE, NodeType.SET]:
            return False
        if node.parent.node_type != NodeType.BAND:
            return False
        return True


@register
class SplitInfo(TransformInfo):
    arg_help = ["Index where the sequence should be split"]

    @staticmethod
    def valid(node: Node):
        if node.node_type not in [NodeType.SEQUENCE, NodeType.SET]:
            return False
        if len(node.children) < 2:
            return False
        if node.parent.node_type != NodeType.BAND:
            return False
        args = SplitInfo.available_args(node)
        if any([isinstance(a, LowerUpperBound) and a.lower >= a.upper for a in args]):
            return False
        return args is not None

    @staticmethod
    def valid_args(node: Node, split_idx: int) -> bool:
        nc = len(node.children)
        return 1 <= split_idx and split_idx < nc - 1

    @staticmethod
    def available_args(node: Node) -> list:
        nc = len(node.children)
        return [LowerUpperBound(lower=1, upper=nc - 1)]


@register
class ScaleInfo(TransformInfo):
    arg_help = []

    @staticmethod
    def valid(node: Node):
        return node.node_type != NodeType.BAND

    @staticmethod
    def available_args(node: Node) -> list:
        return [LowerUpperBound()]


@register
class FullShiftValInfo(TransformInfo):
    arg_help = ["Value"]

    @staticmethod
    def available_args(node: Node):
        return [LowerUpperBound()]


@register
class PartialShiftValInfo(TransformInfo):
    arg_help = ["Statement index", "Value"]

    @staticmethod
    def valid_args(node: Node, stmt_idx: int, value: int):
        return TransformInfo._is_valid_stmt_idx(node, stmt_idx)

    @staticmethod
    def available_args(node: Node):
        ns = len(node.loop_signature)
        return [LowerUpperBound(lower=0, upper=ns), LowerUpperBound()]


@register
class FullShiftVarInfo(TransformInfo):
    arg_help = ["Variable index", "Coefficient"]

    @staticmethod
    def valid(node: Node) -> bool:
        if node.node_type != NodeType.BAND:
            return False
        args = FullShiftVarInfo.available_args(node)
        if not args:
            return False
        return bool(args[0])

    @staticmethod
    def valid_args(node: Node, var_idx: int, _coeff: int):
        args = FullShiftVarInfo.available_args(node)
        if not args:
            return []
        return var_idx in args[0]

    @staticmethod
    def available_args(node: Node):
        if not all(s["vars"] for s in node.loop_signature):
            return []
        # "transpose" loop signatures
        zs = zip(*[s["vars"] for s in node.loop_signature])
        # var is same at idx for all loop_signatures
        same = [len(set(z)) == 1 for z in zs]
        # index of the first "false" in same (or len(same))
        diff_idx = same.index(False) if not all(same) else len(same)
        return [list(range(diff_idx)), LowerUpperBound()]


@register
class PartialShiftVarInfo(TransformInfo):
    arg_help = ["Statement index", "Variable index", "Coefficient"]

    @staticmethod
    def valid_args(node: Node, stmt_idx: int, var_idx: int, coeff: int):
        args = PartialShiftVarInfo.available_args(node)
        if not args:
            return []
        return [stmt_idx, var_idx] in args[0]

    @staticmethod
    def available_args(node: Node):
        args = []
        for stmt_idx, ls in enumerate(node.loop_signature):
            for var_idx in range(len(ls["vars"])):
                args.append([stmt_idx, var_idx])
        return [args, LowerUpperBound()]


@register
class FullShiftParamInfo(TransformInfo):
    arg_help = ["Parameter index", "Coefficient"]

    @staticmethod
    def valid(node: Node) -> bool:
        if node.node_type != NodeType.BAND:
            return False
        args = FullShiftParamInfo.available_args(node)
        if not args:
            return False
        return bool(args[0])

    @staticmethod
    def valid_args(node: Node, param_idx: int, coeff: int):
        args = FullShiftParamInfo.available_args(node)
        if not args:
            return False
        return param_idx in args[0]

    @staticmethod
    def available_args(node: Node):
        if not all(s["params"] for s in node.loop_signature):
            return []
        min_np = len(node.loop_signature[0]["params"])
        return [list(range(min_np)), LowerUpperBound()]


@register
class PartialShiftParamInfo(TransformInfo):
    arg_help = ["Statement index", "Parameter index", "Coefficient"]

    @staticmethod
    def valid(node: Node) -> bool:
        if node.node_type != NodeType.BAND:
            return False
        args = PartialShiftParamInfo.available_args(node)
        if not args:
            return False
        return bool(args[0])

    @staticmethod
    def valid_args(node: Node, stmt_idx: int, param_idx: int, coeff: int):
        args = PartialShiftParamInfo.available_args(node)
        if not args:
            return False
        return [stmt_idx, param_idx] in args[0]

    @staticmethod
    def available_args(node: Node):
        args = []
        for stmt_idx, ls in enumerate(node.loop_signature):
            for param_idx in range(len(ls["params"])):
                args.append([stmt_idx, param_idx])
        return [args, LowerUpperBound()]


@register
class SetParallelInfo(TransformInfo):
    arg_help = ["omp num_threads"]

    @staticmethod
    def available_args(node: Node):
        nproc = multiprocessing.cpu_count()
        return [
            LowerUpperBound(0, nproc),
        ]


@register
class SetLoopOptInfo(TransformInfo):
    arg_help = ["Iterator index", "Option"]

    @staticmethod
    def available_args(node: Node):
        # TODO: I don't know the first element
        return [
            LowerUpperBound(0, 1),
            [
                AstLoopType.DEFAULT.value,
                AstLoopType.ATOMIC.value,
                AstLoopType.SEPARATE.value,
            ],
        ]


@cython.cclass
class Scop:
    ptr_ccscop: cython.pointer(ccScop)

    @staticmethod
    @cython.cfunc
    def create(ptr: cython.pointer(ccScop)) -> Scop:
        scop = Scop()
        scop.ptr_ccscop = ptr
        return scop

    @property
    def schedule_tree(self) -> list[Node]:
        self._goto_root()
        nodes: list[Node] = []
        self._traverse(nodes, parent=-1, location=[])
        return nodes

    @property
    def legal(self) -> bool:
        return bool(self.ptr_ccscop.current_legal)

    def transform_list(self, trs: list) -> list[bool]:
        result = []
        for node_idx, tr, *args in trs:
            node = self.schedule_tree[node_idx]
            result.append(node.transform(tr, *args))
        return result

    def reset(self):
        self.ptr_ccscop.reset()

    def rollback(self) -> None:
        """Roll back (revert) the last transformation."""
        self.ptr_ccscop.rollback()

    def __repr__(self):  # todo this is not a good node representation
        node = self.ptr_ccscop.current_node
        return isl.isl_schedule_node_to_str(node).decode()

    @cython.cfunc
    def _cur(self) -> isl.schedule_node:
        # shortcut to save typing
        return self.ptr_ccscop.current_node

    def _node_type_is(self, node_type: NodeType):
        tmp = isl.isl_schedule_node_get_type(self._cur())
        return NodeType(tmp) == node_type

    def _get_label(self):
        if not self._node_type_is(NodeType.BAND):
            return ""
        mupa = isl.isl_schedule_node_band_get_partial_schedule(self._cur())
        label = isl.isl_multi_union_pw_aff_get_tuple_name(mupa, isl.isl_dim_out)
        mupa = isl.isl_multi_union_pw_aff_free(mupa)
        return label.decode()

    def _get_loop_signature(self) -> str:
        # Scop *si = app->scops[scop_idx];
        if not self._node_type_is(NodeType.BAND):
            return "[]"
        mupa = isl.isl_schedule_node_band_get_partial_schedule(self._cur())
        out_dims = isl.isl_multi_union_pw_aff_dim(mupa, isl.isl_dim_out)
        assert out_dims == 1, SHOULD_NOT_HAPPEN
        # TODO save name??? Comment copied from `times of old'(tm)
        domain = isl.isl_multi_union_pw_aff_domain(mupa)
        num_sets = isl.isl_union_set_n_set(domain)
        slist = isl.isl_union_set_get_set_list(domain)
        result = "["
        for set_idx in range(num_sets):
            if set_idx:
                result += ", "
            set = isl.isl_set_list_get_at(slist, set_idx)
            num_params = isl.isl_set_dim(set, isl.isl_dim_param)
            result += "{'params': ["
            for di in range(num_params):
                if di:
                    result += ", "
                dim_name = isl.isl_set_get_dim_name(set, isl.isl_dim_param, di)
                result += f"'{dim_name.decode()}'"
            result += "], 'vars': ["
            num_vars = isl.isl_set_dim(set, isl.isl_dim_set)
            for di in range(num_vars):
                if di:
                    result += ", "
                dim_name = isl.isl_set_get_dim_name(set, isl.isl_dim_set, di)
                result += f"'{dim_name.decode()}'"
            result += "]}"
            isl.isl_set_free(set)
        result += "]"
        isl.isl_set_list_free(slist)
        isl.isl_union_set_free(domain)
        return result

    def _get_expr(self) -> str:
        if not self._node_type_is(NodeType.BAND):
            return ""
        cur = self.ptr_ccscop.current_node
        mupa = isl.isl_schedule_node_band_get_partial_schedule(cur)
        expr = isl.isl_multi_union_pw_aff_to_str(mupa)
        mupa = isl.isl_multi_union_pw_aff_free(mupa)
        return expr.decode()

    def _goto_root(self) -> None:
        ptr = self.ptr_ccscop.current_node
        ptr = isl.isl_schedule_node_root(ptr)
        self.ptr_ccscop.current_node = ptr

    def _goto_parent(self) -> None:
        ptr = self.ptr_ccscop.current_node
        ptr = isl.isl_schedule_node_parent(ptr)
        self.ptr_ccscop.current_node = ptr

    def _goto_child(self, n: int) -> None:
        ptr = self.ptr_ccscop.current_node
        ptr = isl.isl_schedule_node_child(ptr, n)
        self.ptr_ccscop.current_node = ptr

    def _locate(self, loc: list[int]) -> None:
        # todo (low priority): This might be done in a better way by
        # standardizing accessing the current_node.  We check NULL
        # pointer here because the destructor zeros out the
        # `current_node`, the pointer to ccScop is not zero.  One idea
        # is to clear the pointer in Scop.__dealloc__ (or __del__),
        # but I need to see when is that executed.
        if self.ptr_ccscop.current_node == cython.NULL:
            raise RuntimeError(DELETED_TRANSLATOR_EXCMSG)
        self._goto_root()
        for child in loc:
            self._goto_child(child)

    def _yaml_str(self) -> str:
        node = self.ptr_ccscop.current_node
        return isl.isl_schedule_node_to_str(node).decode()

    def _traverse(self, nodes: list[Node], parent: int, location: list[int]):
        current_idx = len(nodes)
        node = Node.create(self.ptr_ccscop, parent, current_idx, location)
        nodes.append(node)
        if not node.node_type == NodeType.LEAF:
            for c in range(node.num_children):
                self._goto_child(c)
                node.children_idx[c] = len(nodes)
                self._traverse(nodes, current_idx, location + [c])
                self._goto_parent()



import math
import random

from colorama import Fore, Style

from mcts import config


class MCTSNode:
    def __init__(self, parent=None, app=None, action=None, initial_time=None):
        self.parent = parent
        self.app = app
        self.action = action
        self.children = None
        self._number_of_visits = 0
        self.initial_time = initial_time
        self.speedup = -1
        self.is_best = False
        self.best = None

    # TODO: make this config
    def select_child(self):
        return self.select_child_random()

    def select_child_random(self):
        child = random.choice(self.children)
        return child

    # TODO: add policy component
    def select_child_PUCT(self):
        """Implement Upper Confidence Bound sampling strategy"""
        # return random.choice(self.children)
        # TODO: consider softmax
        exploration_constant = 1
        best_score = -1
        best_child = self.children[0]
        for child in self.children:
            if child._number_of_visits == 0:
                return child
            else:
                # UCB
                # UCT(node) = Q(node) + C * sqrt(ln(N(parent))/N(node))
                # Q(node): Average reward (win rate) of the node (exploitation).
                # N(parent): Number of visits to the parent node.
                # N(node): Number of visits to the node.
                # C: Exploration constant (balances exploration vs. exploitation).
                # can start from C=1
                exploitation_term = child.speedup / child._number_of_visits
                # exploration_term = exploration_constant * math.sqrt(math.log(self._number_of_visits) / child._number_of_visits)
                # ucb_score = exploitation_term + exploration_term
                # PUCT
                # adding a component from policy
                policy_child = 1 / len(self.children)
                exploitation_term = policy_child * (
                    math.sqrt(self._number_of_visits) / (1 + child._number_of_visits)
                )
                ucb_score = exploitation_term + exploration_constant * exploitation_term
                if ucb_score > best_score:
                    best_score = ucb_score
                    best_child = child
        # print("SCORES:",scores)
        # child = random.choice(self.children)
        return best_child

    def get_initial_time(self):
        if self.initial_time is not None:
            return self.initial_time
        return self.parent.get_initial_time()

    # def get_networkx_tree(self, graph=None, parent=None):
    #     # TODO: this is WIP
    #     if graph is None:
    #         graph = nx.Graph()
    #     graph.add_node(self)
    #     return graph

    def show_best_source(self):
        if self.best:
            self.best.show_best_source()
        else:
            print()
            print("speedup :", self.speedup)
            print("source  :", self.app.source)

    def set_best(self):
        self.is_best = True
        if self._number_of_visits == 0:
            return
        if self.children:
            best = max(self.children, key=lambda x: x.speedup)
            if best._number_of_visits == 0:
                return
            if best.speedup < self.speedup:
                return
            if (
                self.parent
                and hasattr(self, "evaluate")
                and best.speedup == self.speedup
            ):
                return
            best.is_best = True
            self.best = best
            best.set_best()

    def print(self, depth=0):
        if self._number_of_visits == 0:
            return
        print(f"{' '*depth}", end="")
        if self.is_best:
            print(Fore.GREEN, end="")
        print(f"V:{self._number_of_visits} S:{self.speedup:0.4f} |", self.action)
        print(Style.RESET_ALL, end="")
        if self.children is None:
            return
        for c in self.children:
            c.print(depth + 1)

    def print_best(self, depth=0):
        if self._number_of_visits == 0:
            return
        print(f"{' '*depth}", end="")
        print(f"V:{self._number_of_visits} S:{self.speedup:0.4f} |", self.action)
        if self.best:
            self.best.print_best(depth + 1)

    def update_stats(self, speedup, transforms, source):
        epsilon = 0.1
        if abs(speedup - 1) < epsilon:
            speedup = 1
            # print("QUIT ON ", speedup)
            # return

        if self.speedup is None or speedup > self.speedup:
            self.speedup = speedup
            if self.parent:
                self.parent.update_stats(speedup, transforms, source)
            else:
                self.logger.log(speedup, transforms, source)

    @staticmethod
    def filter_transformations(available_transformations):
        filtered_transformations = []
        if "whitelist_transformations" in config:
            for tr in available_transformations:
                if tr in config["whitelist_transformations"]:
                    filtered_transformations.append(tr)
            return filtered_transformations
        return available_transformations



from tadashi import TrEnum
from tadashi.apps import Simple

app = Simple("./examples/inputs/simple/stencil.c")
app.compile()
print(app.scops[0].schedule_tree[0].yaml_str)

print(f"{app.measure()=}")

app.scops[0].transform_list([[1, TrEnum.TILE1D, 128]])
tile1 = app.generate_code("tile1.c")

tile2 = tile1.generate_code("tile2.c")
tile2.scops[0].transform_list([[3, TrEnum.TILE1D, 16]])

tile2.scops[0].transform_list([[2, TrEnum.INTERCHANGE]])
final_app = tile2.generate_code("final.c")
final_app.compile()
print(f"{final_app.measure()=}")
print("all done")



#!/usb/bin/env python
import functools
import re

import gdb.printing


def get_string(s):
    main = str(s).split()[1:]
    return " ".join(main)[1:-1]


def deref(val):
    return val.dereference() if bool(val) else ""


def sderef(val):
    return str(deref(val))


class PrinterBase:
    def __init__(self, val):
        self.val = val

    def get_array(self, key, nkey, fn, sep=None):
        """Print an array like member.

        Arguments:

            key[str]: The member where the data is.

            nkey[str]: A string of the form 'key1->key2->...->keyN',
                which describes the number of data elements which will
                be read.

            fn[Callable]: The transformation performed on each data
                element.

            sep[Optional[str]]: What is the separator for the
                elements.

        """
        arr = self.val[key]
        cur = self.val
        for k in nkey.split("->"):
            cur = cur[k]
        rng = range(int(cur))
        result = [fn(arr[i]) for i in rng]
        if sep:
            return sep.join(result)
        return result

    def print_members(self, members):
        result = ""
        for member in members:
            if len(member) == 3:
                tag, key, fn = member
                if fn == deref and int(self.val[key]) == 0:
                    continue
                result += f"{tag}:{fn(self.val[key])}; "
            elif len(member) == 5:
                tag, key, nkey, fn, sep = member
                sstr = sep if sep else ""
                result += f"{tag}:{sstr}{self.get_array(key, nkey, fn, sep)}; {sstr}"
            else:
                raise ValueError(f"Bad members: {members}")
        return result


class PlutoProgPrinter(PrinterBase):
    def to_string(self):
        """
        $1 = {
          OK: stmts = 0x555555a97730, nstmts = 2,
          TODO: deps = 0x555555ab7b90, ndeps = 6,
          OK: transdeps = 0x0, ntransdeps = 0,
          OK: data_names = 0x555555a96260, num_data = 3,
          params = 0x555555a96280,
          num_hyperplanes = 3,
          ddg = 0x0,
          fcg = 0x0,
          hProps = 0x555555aaf780,
          nvar = 1,
          npar = 1,
          param_context = 0x555555a91060,
          decls = 0x7ffff3bdb010 "",
          codegen_context = 0x555555a97610,
          globcst = 0x0,
          evicted_hyp_pos = 0,
          is_diamond_tiled = false,
          num_parameterized_loops = -1,
          num_stmts_to_be_coloured = 0,
          scaled_dims = 0x0,
          total_coloured_stmts = 0x0,
          coloured_dims = 0,
          mipTime = 0,
          ilpTime = 0,
          cst_solve_time = 0,
          cst_const_time = 0,
          cst_write_time = 0,
          scaling_cst_sol_time = 0,
          skew_time = 0,
          stencil_check_time = 0,
          fcg_const_time = 0,
          fcg_colour_time = 0,
          fcg_dims_scale_time = 0,
          fcg_update_time = 0,
          fcg_cst_alloc_time = 0,
          tss_time = 0,
          num_lp_calls = 0,
          context = 0x555555a90510
        }
        """
        members = [
            ("stmts", "stmts", "nstmts", sderef, "\n"),  # OK!
            ("deps", "deps", "ndeps", sderef, "\n"),
            ("tdeps", "transdeps", "ntransdeps", sderef, "\n"),  # zero
            ("dnames", "data_names", "num_data", get_string, None),  # OK!
        ]
        statement = self.print_members(members)
        return f"{statement}"


class PlutoAccessPrinter(PrinterBase):
    def to_string(self):
        """
        typedef struct pluto_access {
        int sym_id;
        char *name;

        /* scoplib_symbol_p symbol; */

        PlutoMatrix *mat;
        } PlutoAccess;
        """
        name = get_string(self.val["name"])
        mat = sderef(self.val["mat"])
        return f"{name}{mat}"


class PlutoConstraintPrinter(PrinterBase):
    def to_string(self):
        buf = self.val["buf"]
        ncols = int(self.val["ncols"])
        nrows = int(self.val["nrows"])
        names = self.get_array("names", "ncols", get_string)
        eqs = []
        for row in range(nrows):
            is_eq = "=" if bool(self.val["is_eq"][row]) else "<="
            base = row * int(self.val["alloc_ncols"])
            terms = [f"0{is_eq}"]
            terms += [
                f"{buf[base + i]}{names[i]}" for i in range(ncols) if int(buf[base + i])
            ]
            eqs.append("".join([(t if t[0] == "-" else f"+{t}") for t in terms]))

        return f"Cntr:{eqs}"


class PlutoMatrixPrinter(PrinterBase):
    def to_string(self):
        buf = self.val["val"]
        cols = range(int(self.val["ncols"]))
        rows = range(int(self.val["nrows"]))
        alloc_ncols = int(self.val["alloc_ncols"])
        rows_strs = []
        for row in rows:
            row = [str(buf[row][i]) for i in cols]
            rows_strs.append(",".join(row))
        return f"{{{'|'.join(rows_strs)}}}"


class PlutoHypTypePrinter(PrinterBase):
    def to_string(self):
        typs = ["UNKNOWN", "LOOP", "TILE_SPACE_LOOP", "SCALAR"]
        return typs[int(self.val)]


class StmtPrinter(PrinterBase):
    def to_string(self):
        members = [
            ("id", "id", int),  # OK!
            ("it", "iterators", "dim", get_string, None),  # OK!
            ("dom", "domain", deref),  # OK!
            ("t", "text", get_string),  # OK!
            ("iol", "is_orig_loop", "dim", int, None),  # OK!,
            # ("dim", "dim", int), # OK!
            ("do", "dim_orig", int),  # OK!
            ("t", "tile", int),  # OK!
            ("tr", "trans", deref),  # OK!
            ("eh", "evicted_hyp", deref),  # OK!
            ("ehp", "evicted_hyp_pos", int),  # OK!
            ("hty", "hyp_types", "trans->nrows", str, None),  # OK!
            ("ntl", "num_tiled_loops", int),  # OK!
            ("rx", "reads", "nreads", sderef, None),  # OK!
            # ("nr", "nreads", int),  # OK!
            ("wx", "writes", "nwrites", sderef, None),  # OK!
            # ("nw", "nwrites", int),  # OK!
            ("sid", "scc_id", int),  # OK!
            ("cid", "cc_id", int),  # OK!
            ("ftd", "first_tile_dim", int),  # OK!
            ("ltd", "last_tile_dim", int),  # OK!
            ("typ", "type", int),  # PlutoStmtType NEXT?!
            ("plid", "ploop_id", int),  # OK!
            ("pcs", "parent_compute_stmt", deref),  # OK!
            ("isd", "intra_stmt_dep_cst", int),  # PlutoConstraints*
            ("ps", "pstmt", int),  # pet_stmt*
        ]

        return self.print_members(members)


class DepPrinter(PrinterBase):
    def to_string(self):
        members = [
            ("id", "id", int),  # OK!
            ("s-st", "src", int),  # OK!
            ("d-st", "dest", int),  # OK!
            ("s-ac", "src_acc", sderef),  # OK!
            ("d-ac", "dest_acc", sderef),  # OK!
            ("tmp", "dpolytope", sderef),  # OK!
            # ("tmp", "bounding_poly", sderef),  # OK!
        ]

        return "foobar" + self.print_members(members)


def build_pretty_printer():
    pp = gdb.printing.RegexpCollectionPrettyPrinter("pluto_prog_printer")
    pp.add_printer("PlutoProg", "^plutoProg *$", PlutoProgPrinter)
    pp.add_printer("stmt", "^statement *$", StmtPrinter)
    pp.add_printer("pluto_constraint", "^pluto_constraints *$", PlutoConstraintPrinter)
    pp.add_printer("pluto_matrix", "^pluto_matrix *$", PlutoMatrixPrinter)
    pp.add_printer("pluto_access", "^pluto_access *$", PlutoAccessPrinter)
    pp.add_printer("hyptype", "^hyptype *$", PlutoHypTypePrinter)
    pp.add_printer("dependence", "^dependence *$", DepPrinter)
    return pp
